import { Buffer } from "./buffer.js";
export class PrematureEOF extends Error {
    constructor(...args) {
        super(...args);
    }
}
export class InputStream {
    #onYield;
    /** Set a callback function to be called when the stream is about to
     * suspend for more data. The value returned by the callback will be
     * passed to the `yield` operator.
     */
    onYield(cb) {
        this.#onYield = cb;
    }
    /** Call this method to wait for input.
     */
    *needMore() {
        if (this.#onYield) {
            return yield this.#onYield();
        }
        else {
            throw new TypeError("onYield callback has not been set");
        }
    }
    /** Get an unsigned 8-bits integer without consuming it. */
    *peekUint8() {
        const buf = yield* this.unsafePeek(1);
        return buf.getUint8(0);
    }
    /** Get an unsigned 16-bits integer without consuming it. */
    *peekUint16(littleEndian = false) {
        const buf = yield* this.unsafePeek(2);
        return buf.getUint16(0, littleEndian);
    }
    /** Get an unsigned 32-bits integer without consuming it. */
    *peekUint32(littleEndian = false) {
        const buf = yield* this.unsafePeek(4);
        return buf.getUint32(0, littleEndian);
    }
    /** Consume an unsigned 8-bits integer. */
    *readUint8() {
        const buf = yield* this.unsafePeek(1);
        try {
            return buf.getUint8(0);
        }
        finally {
            buf.drop(1);
        }
    }
    /** Consume an unsigned 16-bits integer. */
    *readUint16(littleEndian = false) {
        const buf = yield* this.unsafePeek(2);
        try {
            return buf.getUint16(0, littleEndian);
        }
        finally {
            buf.drop(2);
        }
    }
    /** Consume an unsigned 32-bits integer. */
    *readUint32(littleEndian = false) {
        const buf = yield* this.unsafePeek(4);
        try {
            return buf.getUint32(0, littleEndian);
        }
        finally {
            buf.drop(4);
        }
    }
}
/** An in-memory input stream that reads from a {@link Buffer} that is
 * given to its constructor. This stream never yields and thus the
 * `onYield` callback will never be called.
 */
export class BufferInputStream extends InputStream {
    #buf;
    constructor(buf) {
        super();
        this.#buf = buf;
    }
    *isEOF() {
        return this.#buf.length <= 0;
    }
    gotData(_input) {
        // Ignore the input.
    }
    *unsafePeek(length) {
        if (this.#buf.length >= length) {
            return this.#buf;
        }
        else {
            throw new PrematureEOF(`Reached the end of buffer before peeking ${length} octets`);
        }
    }
    *read(length) {
        yield* this.unsafePeek(length);
        return this.#buf.take(length);
    }
    *readSome(length) {
        return this.#buf.length > 0
            ? this.#buf.take(length)
            : undefined;
    }
    *skip(length) {
        yield* this.unsafePeek(length);
        this.#buf.drop(length);
    }
}
/** An input stream that gets data from the `yield` operator. Calling
 * `Generator.prototype.next` without an argument signals an EOF. The data
 * is internally buffered in the stream, and its ownership is taken by the
 * stream. Do not mutate it afterwards.
 */
export class PushInputStream extends InputStream {
    #buf;
    #isClosed;
    constructor() {
        super();
        this.#buf = new Buffer();
        this.#isClosed = false;
    }
    *isEOF() {
        if (this.#buf.length > 0) {
            return false;
        }
        else if (this.#isClosed) {
            return true;
        }
        else {
            const input = yield* this.needMore();
            this.gotData(input);
            if (input) {
                return false;
            }
            else {
                this.#isClosed = true;
                return true;
            }
        }
    }
    gotData(input) {
        if (input) {
            if (this.#isClosed) {
                throw new Error("Cannot push data after signaling EOF");
            }
            this.#buf.append(input);
        }
    }
    *unsafePeek(length) {
        while (true) {
            if (this.#buf.length >= length) {
                return this.#buf;
            }
            else {
                const input = yield* this.needMore();
                this.gotData(input);
                if (!input) {
                    throw new PrematureEOF(`Got an EOF before peeking ${length} octets`);
                }
            }
        }
    }
    *read(length) {
        yield* this.unsafePeek(length);
        return this.#buf.take(length);
    }
    *readSome(length) {
        if (this.#buf.length == 0) {
            this.gotData(yield* this.needMore());
        }
        return this.#buf.take(length);
    }
    *skip(length) {
        yield* this.unsafePeek(length);
        this.#buf.drop(length);
    }
}
//# sourceMappingURL=input.js.map
