import { Buffer } from "./buffer.js";
export class OutputStream {
    #onYield;
    /** Set a callback function to be called when the stream is about to
     * suspend for producing data. The value returned by the `yield`
     * operator will be passed to the callback.
     */
    onYield(cb) {
        this.#onYield = cb;
    }
    /** Call this method to produce some output.
     */
    *flush(output) {
        if (this.#onYield) {
            this.#onYield(yield output);
        }
        else {
            throw new TypeError("onYield callback has not been set");
        }
    }
}
/** An in-memory output stream that writes data to a {@link Buffer}. This
 * stream never yields and thus the `onYield` callback will never be
 * called.
 */
export class BufferOutputStream extends OutputStream {
    /** This is the buffer that data will be written to. */
    data;
    constructor() {
        super();
        this.data = new Buffer();
    }
    *write(data) {
        this.data.append(data);
    }
    *writeUint8(n) {
        this.data.appendUint8(n);
    }
    *writeUint16(n, littleEndian = false) {
        this.data.appendUint16(n, littleEndian);
    }
    *writeUint32(n, littleEndian = false) {
        this.data.appendUint32(n, littleEndian);
    }
    takeData() {
        // Nothing to produce.
    }
}
/** An output stream that pushes data to the `yield` operator. The data is
 * internally buffered in the stream.
 */
export class PushOutputStream extends OutputStream {
    #buf;
    #bufSize;
    /** @param bufSize The number of octets to store in the buffer before
     * yielding. 0 means no data will ever be buffered.
     */
    constructor(bufSize) {
        super();
        this.#buf = new Buffer();
        this.#bufSize = bufSize;
        this.#buf.reserve(bufSize);
    }
    *#maybeFlush() {
        if (this.#buf.length >= this.#bufSize) {
            this.flush(this.#buf.toUint8Array());
            this.#buf.clear();
        }
    }
    *write(data) {
        this.#buf.append(data);
        yield* this.#maybeFlush();
    }
    *writeUint8(n) {
        this.#buf.appendUint8(n);
        yield* this.#maybeFlush();
    }
    *writeUint16(n, littleEndian = false) {
        this.#buf.appendUint16(n, littleEndian);
        yield* this.#maybeFlush();
    }
    *writeUint32(n, littleEndian = false) {
        this.#buf.appendUint32(n, littleEndian);
        yield* this.#maybeFlush();
    }
    takeData() {
        try {
            return this.#buf.toUint8Array();
        }
        finally {
            this.#buf.clear();
        }
    }
}
//# sourceMappingURL=output.js.map
