import { LootTable, LootCondition, LootEntry, LootPool, randomIntInClosedInterval } from "../loot-table.js";
/** Mixin for blocks that should be considered equivalent ignoring a
 * certain block state. This mixin overrides `isEquivalentTo`.
 */
export function IgnoringState(base, state) {
    class IgnoringState extends base {
        isEquivalentTo(pa, pb) {
            if (pa.typeId === pb.typeId) {
                for (const [key, value] of pa.states) {
                    if (key === state)
                        continue;
                    else if (pb.states.get(key) !== value)
                        return false;
                }
                return true;
            }
            else {
                return false;
            }
        }
    }
    return IgnoringState;
}
/// Mixin for blocks requiring a silk-touch tool to quick-mine.
export function SilkTouchForQuickMining(base) {
    class SilkTouchForQuickMining extends base {
        isToolSuitable(perm, tool, prefs) {
            if (tool.enchantments.has("silk_touch"))
                return super.isToolSuitable(perm, tool, prefs);
            else
                return false;
        }
    }
    return SilkTouchForQuickMining;
}
/// Mixin for blocks requiring a silk-touch tool to drop itself.
export function SilkTouchForDrop(base) {
    class SilkTouchForDrop extends base {
        lootTable(perm) {
            const stack = perm.getItemStack(1);
            if (stack)
                return new LootTable()
                    .when(LootCondition.matchTool().enchantment("silk_touch"), [new LootPool().entry(stack)]); // 100% drop
            else
                // No corresponding items exist for this block. Drop nothing.
                return new LootTable();
        }
    }
    return SilkTouchForDrop;
}
/** Combination of `SilkTouchForQuickMining` and `SilkTouchForDrop`. */
export function SilkTouchRequired(base) {
    return SilkTouchForQuickMining(SilkTouchForDrop(base));
}
/// Mixin for blocks that yield experience when mined with a non-silk-touch tool.
export function YieldsExperience(base, min, max = min) {
    class YieldsExperience extends base {
        experience(_perm, tool) {
            if (tool.enchantments.has("silk_touch"))
                return 0;
            else
                return randomIntInClosedInterval(min, max);
        }
    }
    return YieldsExperience;
}
/// Mixin for blocks whose drops are affected multiplicatively by Fortune.
export function MultiplicativeDrops(base, block, minRolls, maxRolls, drop) {
    const loots = new LootTable()
        .when(LootCondition.matchTool().enchantment("silk_touch"), [new LootPool().entry(block)]) // 100% drop
        .otherwise([new LootPool()
            .entry(LootEntry
            .item(drop)
            .amount(minRolls, maxRolls)
            .multiplicative())
    ]);
    class MultiplicativeDrops extends base {
        lootTable(_perm) {
            return loots;
        }
    }
    return MultiplicativeDrops;
}
/// Mixin for blocks whose drops use discrete uniform distribution.
export function DiscreteUniformDrops(base, block, minRolls, maxRolls, limit, drop) {
    const loots = new LootTable()
        .when(LootCondition.matchTool().enchantment("silk_touch"), [new LootPool().entry(block)]) // 100% drop
        .otherwise([new LootPool()
            .entry(LootEntry
            .item(drop)
            .amount(minRolls, maxRolls)
            .discreteUniform(limit))
    ]);
    class DiscreteUniformDrops extends base {
        lootTable(_perm) {
            return loots;
        }
    }
    return DiscreteUniformDrops;
}
/// Mixin for blocks requiring stone-tier tools to mine.
export function StoneTier(base) {
    class StoneTier extends base {
        isToolSuitable(perm, tool, prefs) {
            if (tool.tags.has("minecraft:stone_tier") ||
                tool.tags.has("minecraft:iron_tier") ||
                tool.tags.has("minecraft:diamond_tier") ||
                tool.tags.has("minecraft:netherite_tier"))
                return super.isToolSuitable(perm, tool, prefs);
            else
                return false;
        }
    }
    return StoneTier;
}
/// Mixin for blocks requiring iron-tier tools to mine.
export function IronTier(base) {
    class IronTier extends base {
        isToolSuitable(perm, tool, prefs) {
            if (tool.tags.has("minecraft:iron_tier") ||
                tool.tags.has("minecraft:diamond_tier") ||
                tool.tags.has("minecraft:netherite_tier"))
                return super.isToolSuitable(perm, tool, prefs);
            else
                return false;
        }
    }
    return IronTier;
}
/// Mixin for blocks requiring diamond-tier tools to mine.
export function DiamondTier(base) {
    class DiamondTier extends base {
        isToolSuitable(perm, tool, prefs) {
            if (tool.tags.has("minecraft:diamond_tier") ||
                tool.tags.has("minecraft:netherite_tier"))
                return super.isToolSuitable(perm, tool, prefs);
            else
                return false;
        }
    }
    return DiamondTier;
}
//# sourceMappingURL=mixins.js.map
